################################################################################
#
# FIESscript
#
# pipeLine.py
#
# Last update 20/05/2008
#
################################################################################

"""
   This module also contains routines to control the execution of reduction
   tasks. The call to the taskManager can be scheduled in a child thread,
   which will relieve load from the main thread.
"""

# Import external modules

import config
import tasks
import messageLog
import taskManager
import fileUtils
import os
import Queue
import threading


################################################################################

class Busy(Exception):

  """
     Custom error class, raised when the pipeline is busy and cannot accept
     new jobs. Generates no output
  """

  def __init__(self, *args):
    self.args = args


################################################################################

class pipeLine:

  """
     Contains the pipeLine processing module, that takes care of batch
     processing of files.
  """

  def __init__(self):

    # Create current list of tasks to perform
    self.tasklist = []

    # Define initial reduction mode as first one available.
    errstr = self.set_mode(config.options['currentmode'].value)
    if errstr: messageLog.put(errstr)           # (Should never happen)

    # Create an Queue object that will contain the result of child threads
    self.outQueue = Queue.Queue()


  def set_mode(self, mode, newtasklist=False):

    # Set mode parameter
    self.currentmode = mode
    modesandtasks = config.options['reductionmodes'].value

    # Initialize lists of (selected) tasks
    self.tasklist = []
    self.selectedtasks = {}

    try:
      for (taskname, selected) in modesandtasks[self.currentmode]:
	self.tasklist.append( getattr(tasks, taskname) )
	self.selectedtasks[taskname] = selected
    except KeyError:
      errstr = "Reduction mode %s does not exist" % self.currentmode
      return


  def run_singletask(self, taskname):

    try:
      task = getattr(tasks, taskname)
    except:
      messageLog.put_error("Task name %s does not exist" % taskname)
      return

    # OK to start to run the task; send message to log
    messageLog.put('Running "%s"' % task.buttonText)

    ######################################################################
    errorstring = taskManager.runTask(task)
    ######################################################################

    # Errors encountered? If yes, show them!
    if errorstring:
      messageLog.put_error(errorstring)
    else:
      # Everything went smoothly
      pass



  def process_frame(self, frame, inthread=False):

    """
       Prepare a child thread to do the reduction of a single frame. (Will
       run in main thread)
    """

    try:
      if self.pipethread.isAlive():
        # If busy reducing a frame, say so!
        raise Busy
    except AttributeError:
      # Will be raised if pipethread does not exist. This is OK, because
      # is means no reduction is going on
      pass


    # Make sure the input frames exists
    if not os.path.exists(frame):
      messageLog.put_error('Error processing %s' % frame)
      messageLog.put_error('Frame does not exist (anymore?). Skipped!')
      return

    # Put all the selected tasks (by their checkboxes) in the object 'todotasks'
    todotasks = []
    for task in self.tasklist:
      if self.selectedtasks[task.name]:
        todotasks.append(task.name)

    # Run in separate thread if asked to do so
    if inthread:
      # Start a child thread that will do the data 'crunching'
      messageLog.put('Starting pipeline thread', 9)
      self.pipethread = threading.Thread(None, self.run_pipe,
			   args=(frame, self.outQueue, todotasks, self.currentmode))
      self.pipethread.start()
    else:
      # No threading
      self.run_pipe(frame, self.outQueue, todotasks, self.currentmode)



  def run_pipe(self, inframe, outQueue, todotasks, currentmode):

    "Do the reduction a single frame (Can run in child thread)."

    # Prepare local variables
    doneTasks = []
    outframe = ''
    exit = 0

    # Display an initial message in the logging window
    nchars = len('Processing %s' % (inframe))
    messageLog.put('='*nchars)
    messageLog.put('Processing %s' % (inframe))
    messageLog.put('='*nchars)

    # Remember original name of frame
    firstinframe = inframe

    # Split filename into path, base name and extension
    (path, base) = os.path.split(inframe)
    (outfilebase, extn)  = os.path.splitext(base)

    # Remove existing output files with same basename.
    fileUtils.removefiles(os.path.join(config.options['outpath'].value,
                                        '%s_step*.fits' % outfilebase))

    # Define a (new) configuration option that will contain the name
    # of the current reduction mode. This paramter will be referenced by the
    # IRAF-based tasks, and will search for the calibration files in the
    # corresponding directory "tasks/[mode]".
    config.options['currentmode'].set(self.currentmode)

    # Keep track of number of reduction steps
    step = 0

    # Loop over the tasks
    for task in self.tasklist:

      # Increase task counter
      step = step + 1

      # Skip the tasks if it wasn't in 'todotasks' (i.e. it was not selected
      # by the user).
      if not task.name in todotasks:
        messageLog.put("Skipped task %s" % task.name)
	continue

      # Make sure any previous reduction steps required by the current task
      # have been performed before
      for neededTask in task.prereq:
        if not neededTask in doneTasks:
	  # If the reduction cannot continue, make it clear to the user
	  messageLog.put_error('Cannot execute "%s" without "%s"'
	            % (task.buttonText, getattr(tasks, neededTask).buttonText))
          exit = 1 # And make sure we exit the reduction loop (only 'break'
	           # would not be enough here)

      # Exit if needed
      if exit: break

      # Create name of output frame from step counter and step suffix
      if task.suffix:
	outframe = os.path.join(config.options['outpath'].value,
	                        '%s_step%03i_%s.fits' % (outfilebase, step, task.suffix))
      else:
	outframe = None

      # OK to start to run the task; send message to log
      messageLog.put('Running "%s" with %s' % (task.buttonText, os.path.basename(inframe)), 7)

      # Call the taskManager to do the task executing and retrieve possible
      # errors encountered. Any errors occuring while executing the task will
      # never end the reduction interface. Instead, a (descriptive?) error
      # will be shown in the log

      ######################################################################
      errorstring = taskManager.runTask(task, inframe, outframe)
      ######################################################################

      # Errors encountered? If yes, show them!
      if errorstring != None:
        if errorstring != '': messageLog.put_error(errorstring)
 	break
      else:
        # Everything went smoothly
	pass

      # Append task name to the list of successfully finished tasks
      doneTasks.append(task.name)

      # Only if the task we just executed returned an output file name, the
      # next task should use this as its input file name.
      if outframe and os.path.exists(outframe) and task.output:
	inframe = outframe

    # After the loop over all tasks terminates, fill the outQueue, so that the
    # calling routine knows the reduction of the current frame is finished.
    self.outQueue.put((firstinframe, outframe))

    # And bring the good news to the user
    messageLog.put('Finished processing %s' % os.path.basename(firstinframe))



